/*
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	February 1999
 *
 *  This source file is part of the twindex program, and is under the GPL.
 *  Program: twindex -- a program for analysis of X-ray diffraction data
 *  from non-merohedrally twinned crystals.
 *  Author:  Paul D. Boyle
 *  Copyright (C) 2000  Paul D. Boyle
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <math.h>
#include "twindex.h"

void CalculateScatteringVector( double mat[][3], ShelxHKL *refl )
{
    int i;

    refl->s_vec.magnitude = 0.0;

    MatrixMultiplyVector( refl->s_vec.c, mat, refl->hkl );
    
    for( i = 0; i < 3; i++ )
         refl->s_vec.magnitude += refl->s_vec.c[i] * refl->s_vec.c[i];

    refl->s_vec.magnitude = sqrt( refl->s_vec.magnitude );
        
    return;
}

void DifferenceScatteringVector( S_Vector *diff, S_Vector s1, S_Vector s2 )
{
     int i;
     diff->magnitude = 0.0;

     for( i = 0; i < 3; i++ ) {
          diff->c[i] = s1.c[i] - s2.c[i];
          diff->magnitude += diff->c[i] * diff->c[i];
     }
     diff->magnitude = sqrt( diff->magnitude );
}

void SVect2Bisect( double lambda, BisectingAngles *angles, 
                   ShelxHKL *refl )
{
    double mag = 0.0; 

    angles->intensity = refl->f_squared;
  
    angles->theta = RAD2DEG( asin(0.5 * lambda * refl->s_vec.magnitude) );

    mag = sqrt( refl->s_vec.c[0] * refl->s_vec.c[0] +
                refl->s_vec.c[1] * refl->s_vec.c[1] );

    angles->chi = RAD2DEG( atan2(refl->s_vec.c[2], mag) );

    angles->phi = RAD2DEG( atan2(-refl->s_vec.c[0], refl->s_vec.c[1]) );
  
   return;
}

void ReIndexReflection( double inv[][3], ShelxHKL *p, ShelxHKL q )
{
    p->f_squared = q.f_squared;
    p->sigma = q.sigma;

    p->s_vec.c[0] = q.s_vec.c[0];
    p->s_vec.c[1] = q.s_vec.c[1];
    p->s_vec.c[2] = q.s_vec.c[2];

    p->s_vec.magnitude = q.s_vec.magnitude;

    MatrixMultiplyVector( p->hkl, inv, q.s_vec.c );

    return;
}

void GenerateTwinIndicies( double twin_law[][3],
                           ShelxHKL *twin, ShelxHKL pri_frag )
{
    MatrixMultiplyVector( twin->hkl, twin_law, pri_frag.hkl );

    twin->f_squared = pri_frag.f_squared;
    twin->sigma = pri_frag.sigma;
    twin->s_vec = pri_frag.s_vec;

    return;
}

/* TransformOrientMatrices() -- this function takes the primary
 * single domain orientation matrix and the twin law and generates
 * by rotation the orientation matrices of the other n_comp - 1 
 * twin components.
 */
 
void TransformOrientMatrices( PromptData *data )
{
    int i, j;
    double rotate[3][3];
   
/* copy the twin law and then transpose the copy */
    for( i = 0; i < 3; i++ )
         for( j = 0; j < 3; j++ )
              rotate[i][j] = data->twin_law[i][j]; 

    TransposeMatrix( rotate );

/* loop through n_comp - 1 components to generate the orientation
 * matrices for other twin fragments.
 */
    for( i = 0; i < data->n_comp - 1; i++ ) 
         MultiplyMatricies( data->orient_frag[i+1], rotate,
                               data->orient_frag[i] );
     
    return;
}

/* DeriveTwinLaw() -- takes the orientation matricies of the first two 
 * fragments and derives a twin law which relates H' = T * H 
 */

void DeriveTwinLaw(  PromptData *data )
{

    double det;
    double inverted[3][3];

    det = Determinant( data->orient_frag[0] );
    InvertMatrix( det, data->orient_frag[0], inverted );
    MultiplyMatricies( data->twin_law, data->orient_frag[1], inverted );
    TransposeMatrix( data->twin_law );

    return;
}


int RoundToNearestInteger( double num ) 
{
    double fraction, 
           whole;
    int    ret_val;

    fraction = modf( num, &whole );

    if( 0.0 > fraction ) {
        if( -0.5 > fraction ) 
             ret_val = (int)whole - 1;
        else
             ret_val = (int)whole;
    }
    else {
       if( 0.5 > fraction ) 
           ret_val = (int)whole;
       else
           ret_val = (int)whole + 1;
    }
    return ret_val;
}

/* multiples a matrix, M, by a vector V to give product VP as
 * VP = M * V
 */
void MatrixMultiplyVector( double prod[3], double mat[][3], double vec[3] )
{
    int i, j;
    for( i = 0; i < 3; i++ ) {
         prod[i] = 0.0;
         for( j = 0; j < 3; j++ )
              prod[i] += mat[i][j] * vec[j];
    }     
       
}

void MultiplyMatricies( double prod[][3], double a[][3], double b[][3] )
{
    int i, j, k;
    for( i = 0; i < 3; i++ )
         for( j = 0; j < 3; j++ ) {
              prod[i][j] = 0.0;
              for( k = 0; k < 3; k++ )
                   prod[i][j] += a[i][k] * b[k][j];
         }
}

/* Calculates the dot product of 2 vectors. */

double DotProduct( double a[], double b[])
{
    double c;
    int i, NMAX;
    NMAX = sizeof a / sizeof a[0];

    c = 0;
    for( i = 0; i < NMAX; i++ ) 
                c +=  a[i] * b[i];
    
    return c;
}

/* Determinant() */
/* This is a function which calculates the determinant 
 * for a 3x3 matrix.  Nothing fancy here. 
 */
 
double Determinant( double m[][3] )
{
    double determ;
    
    determ = m[0][0] * (m[1][1] * m[2][2] - m[1][2] * m[2][1]) +
             m[1][0] * (m[2][1] * m[0][2] - m[0][1] * m[2][2]) +
             m[2][0] * (m[0][1] * m[1][2] - m[0][2] * m[1][1]);
             
    return determ;
}             

/* InvertMatrix() */
void InvertMatrix( double det, double m[][3], double inv[][3])
{
    inv[0][0] =   (m[1][1] * m[2][2] - m[1][2] * m[2][1]) / det;
    inv[0][1] = - (m[0][1] * m[2][2] - m[0][2] * m[2][1]) / det;
    inv[0][2] =   (m[0][1] * m[1][2] - m[0][2] * m[1][1]) / det;
    inv[1][0] = - (m[1][0] * m[2][2] - m[2][0] * m[1][2]) / det;
    inv[1][1] =   (m[0][0] * m[2][2] - m[0][2] * m[2][0]) / det;
    inv[1][2] = - (m[0][0] * m[1][2] - m[0][2] * m[1][0]) / det;
    inv[2][0] =   (m[1][0] * m[2][1] - m[2][0] * m[1][1]) / det;
    inv[2][1] = - (m[0][0] * m[2][1] - m[0][1] * m[2][0]) / det;
    inv[2][2] =   (m[0][0] * m[1][1] - m[0][1] * m[1][0]) / det;
    
    return;
}

/* TransposeMatrix() -- yep, that's is what it does */
void TransposeMatrix( double m[3][3] )
{
    double tmp;

    int i, j;

    for( i = 1; i < 3; i++ )
       for( j = 0; j < i; j++ ) {
            tmp = m[i][j];
            m[i][j] = m[j][i];
            m[j][i] = tmp;
       }    

}   

/* ExtractMetricTensors() -- takes an orientation matrix and 
 * passes the both the direct and reciprocal metric tensors back
 * to the caller.
 */

void ExtractMetricTensors( double orient[][3],
                           double direct[][3],
                           double recip[][3] )
{
    double tmp[3][3],
           determ;

    int i, j;

/* initialize the direct and recip arrays and copy the orientation matrix
 * into the tmp variable.
 */
    for( i = 0; i < 3; i++ )
         for( j = 0; j < 3; j++ ) {
              tmp[i][j] = orient[i][j];
              direct[i][j] = recip[i][j] = 0.0;
         }

/* transpose orientation matrix because orient * (orient)Transpose =
 * reciprocal metric tensor.
 */
    
    TransposeMatrix( tmp );
    MultiplyMatricies( recip, tmp, orient );

/* invert to get the direct space metric tensor. */

    determ = Determinant( recip );
    InvertMatrix( determ, recip, direct );
    
    return;
}

/* ExtractUnitCell() -- function takes a metric tensor and returns the
 * corresponding cell constants and volume.  The is function has been
 * verified.
 */

void ExtractUnitCell( UnitCell *cell, double m_tensor[3][3] )
{
    double determ = Determinant( m_tensor );
    double ab, ac, bc;

    cell->a = (float)sqrt( m_tensor[0][0] );
    cell->b = (float)sqrt( m_tensor[1][1] );
    cell->c = (float)sqrt( m_tensor[2][2] );
    cell->volume = sqrt( determ );
     
    ab = m_tensor[0][1] / (cell->a * cell->b );
    ac = m_tensor[0][2] / (cell->a * cell->c );
    bc = m_tensor[1][2] / (cell->b * cell->c );

    cell->alpha = RAD2DEG( acos(bc) );
    cell->beta  = RAD2DEG( acos(ac) );
    cell->gamma = RAD2DEG( acos(ab) );

    return;
}
